package filter;
import java.sql.*;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Collections;

public class Filter {

//	Those parameters contain the details of the database connection and the connection itself.
	String database;
	String username;
	String password;
	Connection connection;
	String tablename;
	Shredder shredder;
		
	/**
	 * Making default constructor private in order avoid broken objects with no database connection
	 * @param args
	 */
	private Filter(){};

	/**
	 * This is the actual constructor to use. It automatically connects the class to the given database.
	 * @param database 
	 * @param username
	 * @param password
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	public Filter(String database, String username, String password, String delimiter) throws SQLException, ClassNotFoundException
	{
		this.database=database;
		this.username=username;
		this.password=password;
		this.shredder=new Shredder(delimiter);
				
		//	 Loading database driver
		Class.forName( "com.mysql.jdbc.Driver" );
		
		//	Connecting to the database
		connection = DriverManager.getConnection( database, username, password );
	}
	
	/**
	 * 
	 *  @param delimiter - This is used to set a REGEX for the delimiter that tears apart the input text sources.
	 */
	public void setDelimiter(String delimiter){
		shredder.setDelimiter(delimiter);
	}
			
	/**
	 * @throws SQLException
	 */
	public void sanitizeDatabase() throws SQLException{
		Statement stmt = connection.createStatement();
		stmt.execute("TRUNCATE TABLE bayesfilter");
		stmt.execute("TRUNCATE TABLE bayes_counts");
		stmt.execute("INSERT INTO bayes_counts VALUES(0,0);");
		
		stmt.close();
	}

	/**
	 * 
	 * @param sourceText
	 * @return Returns a Iterator of Strings representing the distinct tokens inside the sourceText 
	 */
	public Iterator derriveTokensFromSource(String sourceText){
		return shredder.shred(sourceText);
	}
		
	/**
	 * 
	 * @param sourceText
	 * @param typeOfInput
	 * @return Interger - Return the number of learned Tokens that have not been seen before
	 */
	public int trainFilter(String sourceText, String typeOfInput){
		ResultSet rSet;
		String sqlQuery="Not set! This shouldn't happen !!";
		Iterator tokensToLearnFrom = derriveTokensFromSource(sourceText);
		int newTokensLearned=0;
				
	    try
	    {
	      String currentToken;
	      Statement stmt = connection.createStatement();
	      PreparedStatement isItThere = connection.prepareStatement("SELECT token from bayesfilter where token=?");
	      PreparedStatement insertIt = connection.prepareStatement("INSERT into bayesfilter (token,"+typeOfInput+") VALUES (?, 1)");
	      PreparedStatement updateIt = connection.prepareStatement("UPDATE bayesfilter SET "+typeOfInput+"="+typeOfInput+"+1 where token=?");
	      	      
	      while(tokensToLearnFrom.hasNext()){
	      
		      //get one token at a time to be added to the statistics
	    	  currentToken=tokensToLearnFrom.next().toString();
	   	  		      
		      //Try to fetch this token from DB
	    	  isItThere.setString(1,currentToken);
	    	  rSet=isItThere.executeQuery();
		      
		      //Token does not exist in DB yet
		      if (!rSet.first()){
		    	  insertIt.setString(1, currentToken);
		    	  insertIt.execute();
		    	  newTokensLearned++;
		      }
		      //Token is already in DB, updating number of matches
		      else{
		       	  updateIt.setString(1, currentToken);
		       	  updateIt.execute();
		      }
		         		      
		  }
	      
	      
	      isItThere.close();
	      insertIt.cancel();
	      updateIt.close();
	      //increase number of total inputs scanned
	      stmt.executeUpdate( "UPDATE bayes_counts SET "+typeOfInput+"="+typeOfInput+"+1;" );
	      stmt.close();
	      }
	    catch ( SQLException e )
	    {
	      System.out.println( "Error accessing database. Query was:"+ sqlQuery + e.getMessage() );
	      return -1;
	    }
	return newTokensLearned;
	}
	
	
	/**
	 * 
	 * @param sourceText Input to be evaluated for being spam
	 * @param numOfTokens Maximum number of tokens to be taken as measurement, most interesting tokens are taken
	 * @return Percentage 
	 */
	public double evaluateInput(String sourceText, int numOfTokens){
		
		if(isDatabaseEmpty())return 0.5;
		else{
						
			Iterator tokenContainersFromDB = queryFilter(sourceText);
			TokenContainer tempcontainer;
						
			int counter=0;
			double aggSpammyness=1.0, aggHammyness=1.0;
			
			while(tokenContainersFromDB.hasNext() && counter!=numOfTokens){
				counter++;
				tempcontainer = (TokenContainer)tokenContainersFromDB.next();
				aggSpammyness*=tempcontainer.getSpammyness();
				aggHammyness*=1.0-tempcontainer.getSpammyness();
			}
					
			double prob=aggSpammyness/(aggSpammyness+aggHammyness);
			return prob;
		}
	}
	
	/**
	 * 
	 * @param tokens An Iterator of Strings containing the tokens to query for
	 * @return Iterator with TokenContainers that are filled with the correct statistical data
	 */
	private Iterator queryFilter(Iterator tokens){
		
		ResultSet rSet;
		String sqlQuery="Not set! This shouldn't happen !!";
		LinkedList tokenContainerList = new LinkedList();
		//Checking if DB is empty
		if (isDatabaseEmpty()) return tokenContainerList.iterator();
		
		else{
		    try
		    {
		    	sqlQuery="select token, spam, ham, spamqot/(spamqot+hamqot) AS spammyness from (select token,spam,ham,(spam / (SELECT spam from bayes_counts)) AS spamqot, (ham / (SELECT ham from bayes_counts)) AS hamqot from bayesfilter WHERE token=?) AS details";
		    	PreparedStatement prepStatement = connection.prepareStatement(sqlQuery);
		    		  
		    	while(tokens.hasNext()){

		    		prepStatement.setString(1, tokens.next().toString());	     	     
			     rSet = prepStatement.executeQuery();
			     
			     if(rSet.first()){
			    	 //System.out.println("Spammyness:  " + rSet.getString("token"));
			    	 tokenContainerList.add(new TokenContainer(rSet.getString("token"), rSet.getInt("spam"), rSet.getInt("ham"), rSet.getDouble("spammyness")));
			     }
			     		     
		      }
			     prepStatement.close();
		      }
		    catch ( SQLException e )
		    {
		      System.out.println( "Error accessing database during queryFilter. Query was:"+ sqlQuery + e );
		    }
			
		    Collections.sort(tokenContainerList);
		    
			return tokenContainerList.iterator();
		}
	}
	
	/**
	 * 
	 * @param sourceText Text to tokenize and then look up in the database as single tokens.
	 * @return Iterator containing TokenContainer objects. The Iterator itself is sorted descending i.e. most important token first.
	 */
	public Iterator queryFilter(String sourceText){
		return queryFilter(shredder.shred(sourceText));
	}

	/**
	 * 
	 * @param numOfTokensToGet Number of the most interesting tokens to get. If you put "-1" here all tokens in the db are returned.
	 * @return Iterator with TokenContainer objects containing the statistical data of the token.
	 */
	public Iterator queryDBForInterestingTokens(int numOfTokensToGet){
				
		ResultSet rSet;
		LinkedList tokenList = new LinkedList();
		
		//Checking if DB is empty
		if (isDatabaseEmpty()) return tokenList.iterator();
		else{
		
		
			//Create sqlQuery with optional Limit
			String sqlQuery="select token, spam, ham, spamqot/(spamqot+hamqot) AS spammyness, abs(0.5-(spamqot/(spamqot+hamqot))) AS interestingness from (select token, spam, ham,(spam / (SELECT spam from bayes_counts)) AS spamqot, (ham / (SELECT ham from bayes_counts)) AS hamqot from bayesfilter) AS getquot order by 5 desc";
			if (numOfTokensToGet>-1)sqlQuery+=" limit "+numOfTokensToGet+";";
			else sqlQuery+=";";
			
			
			
		    try
		    {
		    	Statement stmt = connection.createStatement();
		    	rSet=stmt.executeQuery(sqlQuery); 
		    	    	    		   	     	  
		    	while(rSet.next()){
		    		tokenList.add(new TokenContainer(rSet.getString("token"), rSet.getInt("spam"), rSet.getInt("ham"), rSet.getDouble("spammyness")));
			    }
			     		     
		      }
			       
		    catch ( SQLException e )
		    {
		      System.out.println( "Errors during DB-query of most interesing tokens. Query was:"+ sqlQuery + e );
		    } 
	
		    //Get the details for the tokens and return those TokenContainers as an Iterator
		    return tokenList.iterator(); 
		}	
	}

	/**
	 * 
	 * @return integer arraay with spam count as first and ham count as second entry
	 */
	public int[] queryDBForTotalCounts(){
		ResultSet rSet;
		String sqlQuery="SELECT spam, ham FROM bayes_counts;";
		int number[] = new int[2];	
	    try
	    {
	    	Statement stmt = connection.createStatement();
	    	rSet=stmt.executeQuery(sqlQuery); 
	    	    	    		   	     	  
	    	rSet.first();
	    	number[0]=rSet.getInt(1);
	    	number[1]=rSet.getInt(2);
		     		     
	      }
		       
	    catch ( SQLException e )
	    {
	      System.out.println( "Errors during DB-query of most interesing tokens. Query was:"+ sqlQuery + e );
	    } 
	    
	    return number; 
	}

	/**
	 * 
	 * @return true if DB doesn't have a count in both spam or ham
	 */
	public boolean isDatabaseEmpty(){
		int[] dbcounts = queryDBForTotalCounts();
		if (dbcounts[0]==0&&dbcounts[1]==0) return true;
		else return false;
	}
	
}


